// -*- mode:c++ -*-

#ifndef _KYOKUMEN_H_
#define _KYOKUMEN_H_

#include "koma_inf.h"
#include <stdio.h>   // FILE
#include <stdint.h>  // int64_t, uint64_t


#define DIRECTION_SIZE  12

extern const int Direct[DIRECTION_SIZE];
extern const int CanPromote[];
extern const int CanMove[DIRECTION_SIZE][64];
extern const int CanJump[DIRECTION_SIZE][64];
extern const char* komaStr[];
extern const char* komaStr2[];


// ３章で追加。駒の価値と持ち駒の価値
extern const int KomaValue[];
extern const int HandValue[];

/** 利きビットマスク.
    利きは、８方向の利き＋桂馬の４方向の利き＋飛び駒の８方向の利きを扱います。
    短い攻撃のために、下12ビットに1ビットずつ割り当てます。
    演算の高速性のために、飛び駒の利きは上位１６ビットのうちの下８ビットを利用します。
*/
typedef uint32_t Kiki;

// 5章で追加。６４ビット型の整数型
typedef int64_t int64;
typedef uint64_t uint64;



class Te;

/** 局面のクラス.
    盤 + 持ち駒 + 手番、を持つ。
 */
class Kyokumen
{
public:
  // 桂馬の利きがbanからはみ出すので、はみ出す分を確保しておきます。
  // C++では、構造体の内部の変数の並び順は宣言した順になることを利用しています。
  // 普通はあまり使わない「汚い」テクニックですけど、こういうテクニックもあるということで。
  KomaInf banpadding[16];
  
  /** 盤上の駒.
      [0x11] から [x99] がそのまま１一から９九に対応します。
      駒の位置は、例えば２三なら、(2 * 16 + 3) のようにあらわします。
      ISSUE: padding が少ないため、相対位置を1次元のインデックスで一意に表すことができない。

      2次元配列を使うと遅いので、１次元配列を使います。
      また、掛け算の際に、＊９とかを用いるよりも、2の階乗を掛け算に使うと掛け算が
      早くなるので、＊１６を使います。
  */
  KomaInf ban[16 * 11];

  /** 持ち駒.
      王が持ち駒になることはないので、EHIまでで十分です。
      Hand[SFU] が１なら先手の持ち駒に歩が１枚、Hand[EKI]が３なら敵の持ち駒に金が３枚
      という要領です。
  */
  int Hand[EHI + 1];

  /**この局面の手数です. 
     千日手判定のため、棋譜の局面（ハッシュ）を更新するので、手番ではなく手数を保持します。
  */
  int Tesu;

  /** 駒の利き. 敵の駒と自分の駒の利きは別々に保持します。 
      ISSUE: pin情報を保持せず、MakeLegalMoves() で毎回作り直すため、全般的にコードが冗長。
   */
  Kiki controlS[16 * 11];
  Kiki controlE[16 * 11];

  /** 互いの玉の位置です。 */
  int kingS;
  int kingE;

  /** 局面ハッシュの種. 5章で追加。 */
  static uint64 HashSeed[ERY + 1][0x99 + 1];
  static uint64 HandHashSeed[EHI + 1][18 + 1];

  /** ハッシュ値 */
  uint64 KyokumenHashVal;
  uint64 HandHashVal;
  uint64 HashVal;

  // ３章で追加。局面の評価値
  int value;


public:
  Kyokumen() { }

  /** 盤面と持ち駒から局面を生成するコンストラクタ. ２章で追加。
   */
  Kyokumen(int tesu, const KomaInf ban[9][9], const int Motigoma[]);

  /** 合法手をすべて生成する.
      @return 合法手の数
      ISSUE: Teの配列に格納している。インスタンスのコピーが多く発生。
   */
  int MakeLegalMoves(int SorE, Te* tebuf, int* pin = NULL);  

  /** 王手のみを生成. 第９章で追加。
      @return 王手（生成した手）の数
  */
  int MakeChecks(int SorE, Te* tebuf, int* pin = NULL);

  /** 実際に１手動かす. 2章で追加。 */
  void Move(int SorE, const Te& te);

  int IsLegalMove(int SorE, Te& te);
  

protected:
  // １章で追加。controlS,controlEの初期化。
  void InitControl();
	
  // １章で追加。駒の正しい動きの生成をするための関数群
  void MakePinInf(int *pin) const;

  // １章で追加。正しく駒を動かすための関数群
  int Utifudume(int SorE,int to,int *pin);
  
  int AntiCheck(int SorE,Te *tebuf,int *pin,Kiki control);

  //玉の動く手の生成
  void MoveKing(int SorE,int &teNum,Te *teTop,Kiki kiki);

  //toに動く手の生成
  void MoveTo(int SorE,int &teNum,Te *teTop,int to,int *pin);		

  //toに駒を打つ手の生成
  void PutTo(int SorE,int &teNum,Te *teTop,int to,int *pin);		

  Kiki CountControlS(int pos) const;
  Kiki CountControlE(int pos) const;
  Kiki CountMove(int SorE,int pos,int *pin);

  void AddMoves(int SorE,int &teNum,Te *teTop,int from,int pin,int Rpin=0);

  void AddStraight(int SorE,int &teNum,Te *teTop,int from,int dir,int pin,int Rpin=0);

  void AddMove(int SorE,int &teNum,Te *teTop,int from,int diff,int pin,int Rpin=0);

  // ３章で追加。交換値を求めるための関数群
  int IsCorrectMove(Te &te);
  int EvalMin(Te *MoveS,int NumMoveS,Te *MoveE,int NumMoveE);
  int EvalMax(Te *MoveS,int NumMoveS,Te *MoveE,int NumMoveE);
  int Eval(int pos);

 
public:
  static void HashInit();


  // 1000手くらいしか指さないことにして置きましょう。
  // ISSUE: スレッドごとに持たないと千日手を誤判定する。
  static uint64 HashHistory[1000];
  static int OuteHistory[1000];

  // １章で追加。posからdir方向にある何か駒（もしくは壁）を探す。
  inline int search(int pos, int dir) const {
    do {
      pos += dir;
    } while(ban[pos] == EMPTY);
    return pos;
  }


  /** 画面へ局面を印字する. ２章で追加。 */
  void Print() const {
    FPrint(stdout);
  }

  /** ファイルに局面を印字する. ２章で追加。 */
  void FPrint(FILE* fp) const;


  // ３章で追加。手番側の一番良い交換値を探して返す。
  int BestEval(int SorE);


  int operator == (const Kyokumen& k) const 
  {
    int ret = true;
    for (int suji = 0x10; suji <= 0x90; suji += 0x10) {
      for (int dan = 1; dan <= 9; dan++) {
        if (ban[suji+dan] != k.ban[suji+dan]) {
	  ret=false;
	  break;
	}
      }
    }
    return ret;
  }

  void Dump() const;

  // 第９章で追加。詰め将棋ルーチン。
  int Mate(int SorE,int maxDepth,Te &te);
  int CheckMate(int SorE,int depth, int depthMax, Te *checks,Te &te);
  int AntiCheckMate(int SorE,int depth, int depthMax, Te *checks);
};

#endif
